// DBConnectionManager.java
//package database;
import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.Date;

public class DBConnectionManager {
    static private DBConnectionManager instance;       
    static private int clients;
    
    private long timeout;
    private Vector drivers = new Vector();
    private PrintWriter log;
    private Hashtable pools = new Hashtable();
    
    // Singleton & Double-Checked Locking
    static public DBConnectionManager getInstance() {
        if (instance == null) {
            synchronized(DBConnectionManager.class) {
                if (instance == null) {
                    instance = new DBConnectionManager();
                }
            }
        }
        clients++;
        return instance;
    }

    /**
     * ##
     **************************************************************************/
    static public void refresh() {
        synchronized(DBConnectionManager.class) {
            instance = null;
            clients = 0;
            DBConnectionManager.getInstance();
        }
    }    

    private DBConnectionManager() {
        init();
    }
    
    public void freeConnection(String name, Connection con) {
        DBConnectionPool pool = (DBConnectionPool) pools.get(name);
        if (pool != null) {
            pool.freeConnection(con);
        }
    }

    /**
     * 
     **************************************************************************/
    public Connection getConnection(String name) 
        throws SQLException {
        Connection con;
        DBConnectionPool pool = (DBConnectionPool) pools.get(name);
	System.out.println("************************************ " );
	System.out.println("poolName : " + name);
	System.out.println("pool : " + pool);
	System.out.println("************************************ " );
        if (pool == null) {
            throw new SQLException("pool alias does not exist in file [db.properties]");
        } else {
            con = pool.getConnection();
            if (con == null)
                throw new SQLException("can't get connection from pool [" + name + "]");
                
        }
        return con;
    }
    /*        
    public Connection getConnection(String name) {
        DBConnectionPool pool = (DBConnectionPool) pools.get(name);
        if (pool != null) {
            return pool.getConnection(5000);
        }
        return null;
    }
    */
    
    
    public Connection getConnection(String name, long time) {
        DBConnectionPool pool = (DBConnectionPool) pools.get(name);
        if (pool != null) {
            return pool.getConnection(time);
        }
        return null;
    }
    
    public synchronized void release() {
        if (--clients != 0) {
            return;
        }
        
        Enumeration allPools = pools.elements();
        while (allPools.hasMoreElements()) {
            DBConnectionPool pool = (DBConnectionPool) allPools.nextElement();
            pool.release();
        }
        Enumeration allDrivers = drivers.elements();
        while (allDrivers.hasMoreElements()) {
            Driver driver = (Driver) allDrivers.nextElement();
            try {
                DriverManager.deregisterDriver(driver);
                log("Deregistered JDBC driver " + driver.getClass().getName());
            }
            catch (SQLException e) {
                log(e, "Can't deregister JDBC driver: " + driver.getClass().getName());
            }
        }
    }
    
    private void createPools(Properties props) {
        Enumeration propNames = props.propertyNames();
        while (propNames.hasMoreElements()) {
            String name = (String) propNames.nextElement();
            if (name.endsWith(".url")) {
                String poolName = name.substring(0, name.lastIndexOf("."));
                String url = props.getProperty(poolName + ".url");
                if (url == null) {
                    log("No URL specified for " + poolName);
                    continue;
                }
                String user = props.getProperty(poolName + ".user");
                String password = props.getProperty(poolName + ".password");
                String maxconn = props.getProperty(poolName + ".maxconn", "5");
                String initconn = props.getProperty(poolName + ".initconn", "2");
                int max;
                int init;
                try {
                    max = Integer.valueOf(maxconn).intValue();
                    init = Integer.valueOf(initconn).intValue();
                }
                catch (NumberFormatException e) {
                    log("Invalid maxconn/initconn value " + maxconn + "/" + initconn + " for " + poolName);
                    max = 0;
                    init = 0;
                }
                DBConnectionPool pool = 
                    new DBConnectionPool(poolName, url, user, password, max, init);
			System.out.println("******************** " );  // delete after test
			System.out.println("pool Name : " + poolName);  // delete after test
			System.out.println("pool created : " + pool);  // delete after test
			System.out.println("pool url : " + url);  // delete after test
			System.out.println("pool usr : " + user);  // delete after test
			System.out.println("pool passwprd : " + password);  // delete after test
			System.out.println("pool url : " + url);  // delete after test
			System.out.println("******************** " );  // delete after test
                pools.put(poolName, pool);
                log("Initialized pool " + poolName);
            }
        }
    }
    
    private void init() {
        InputStream is = getClass().getResourceAsStream("db.properties");
        Properties dbProps = new Properties();
        try {
            dbProps.load(is);
        }
        catch (Exception e) {
            System.err.println("Can't read the properties file. " +
                "Make sure db.properties is in the CLASSPATH");
            return;
        }
        timeout = Long.parseLong(dbProps.getProperty("timeout", "5000"));
        String logFile = dbProps.getProperty("logfile", "DBConnectionManager.log");
        try {
            log = new PrintWriter(new FileWriter(logFile, true), true);
        }
        catch (IOException e) {
            System.err.println("Can't open the log file: " + logFile);
            log = new PrintWriter(System.err);
        }

	    dbProps.list(System.out);

        loadDrivers(dbProps);
        createPools(dbProps);
    }
    
    private void loadDrivers(Properties props) {
        String driverClasses = props.getProperty("drivers");
        StringTokenizer st = new StringTokenizer(driverClasses);
        while (st.hasMoreElements()) {
            String driverClassName = st.nextToken().trim();
            try {
                Driver driver = (Driver) 
                    Class.forName(driverClassName).newInstance();
                DriverManager.registerDriver(driver);
                drivers.addElement(driver);
                log("Registered JDBC driver " + driverClassName);
            }
            catch (Exception e) {
                log("Can't register JDBC driver: " +
                    driverClassName + ", Exception: " + e);
            }
        }
    }
    
    private void log(String msg) {
        log.println(new Date() + ": " + msg);
    }
    
    private void log(Throwable e, String msg) {
        log.println(new Date() + ": " + msg);
        e.printStackTrace(log);
    }

    /**
     * ##
     **************************************************************************/
    public void print(Writer out) {
        print(new PrintWriter(out));
    }
    
    public void print(PrintWriter out) {
        out.println("<h3>--DBConnectionPool--</h3>");
        
        out.println("clients     : " + clients + "<br>");
        out.println("timeout     : " + timeout + "<br>");
        Enumeration e = pools.elements();
        while (e.hasMoreElements()) {
            ((DBConnectionPool) e.nextElement()).print(out);
        }
    }

    public void print(PrintStream out) {
        out.print("--DBConnectionPool--\n");
        
        out.print("clients     : " + clients + "\n");
        Enumeration e = pools.elements();
        while (e.hasMoreElements()) {
            ((DBConnectionPool) e.nextElement()).print(out);
        }
    }
    
    /**
     *
     ****/
    class DBConnectionPool {
        private int checkedOut;
        private Vector freeConnections = new Vector();
        private int maxConn;
        private int initConn;
        private String name;
        private String password;
        private String URL;
        private String user;
        
        public DBConnectionPool(String name, String URL, String user, String password, 
                int maxConn, int initConn) {
            this.name = name;
            this.URL = URL;
            this.user = user;
            this.password = password;
            this.maxConn = maxConn;
            this.initConn = initConn;
            
            for (int i=0; i<initConn; i++) {
                Connection conn = newConnection();
                freeConnection(conn);
            }
        }
        
        public synchronized void freeConnection(Connection con) {
            // ##
            try {
                if (con != null && !con.isClosed()) {
                    con.rollback();
                    con.clearWarnings();
                    
                    freeConnections.addElement(con);
                    checkedOut--;
                    notifyAll();
                }
            } catch (SQLException e) {
                log("Exception during freeConnection");
                log(e.toString());
            }
        }
        
        public synchronized Connection getConnection() {
            Connection con = null;
            if (freeConnections.size() > 0) {
                // Pick the first Connection in the Vector
                // to get round-robin usage
                con = (Connection) freeConnections.firstElement();
                freeConnections.removeElementAt(0);
                try {
                    // ##
                    if (con.isClosed() || con==null) {
                        log("Removed bad connection from " + name);
                        // Try again recursively
                        con = getConnection();
                    }
                } catch (SQLException e) {
                    log("Removed bad connection from " + name);
                    // Try again recursively
                    con = getConnection();
                }
            }
            else if (maxConn == 0 || checkedOut < maxConn) {
                con = newConnection();
            }
            if (con != null) {
                checkedOut++;
            }
            return con;
        }
        
        public synchronized Connection getConnection(long timeout) {
            long startTime = new Date().getTime();
            Connection con;
            while ((con = getConnection()) == null) {
                try {
                    wait(timeout);
                }
                catch (InterruptedException e) {}
                if ((new Date().getTime() - startTime) >= timeout) {
                    // Timeout has expired
                    return null;
                }
            }
            return con;
        }
        
        public synchronized void release() {
            Enumeration allConnections = freeConnections.elements();
            while (allConnections.hasMoreElements()) {
                Connection con = (Connection) allConnections.nextElement();
                try {
                    con.close();
                    log("Closed connection for pool " + name);
                }
                catch (SQLException e) {
                    log(e, "Can't close connection for pool " + name);
                }
            }
            freeConnections.removeAllElements();
        }
        
        private Connection newConnection() {
            Connection con = null;
            try {
                if (user == null) {
                    con = DriverManager.getConnection(URL);
                }
                else {
                    con = DriverManager.getConnection(URL, user, password);
                }
                log("Created a new connection in pool " + name);
            }
            catch (SQLException e) {
                log(e, "Can't create a new connection for " + URL);
                return null;
            }
            return con;
        }
        /**
         * 
         **************************************************************************/
        public void print(Writer out) {
            print(new PrintWriter(out));
        }
        
        public void print(PrintWriter out) {
            out.println("------------------------------------------<br>");
            out.println("name       : " + name + "<br>");
            out.println("url        : " + URL + "<br>");
            out.println("user       : " + user + "<br>");
            out.println("password   : " + password + "<br>");
            out.println("maxconn    : " + maxConn + "<br>");
            out.println("initconn   : " + initConn + "<br>");
            out.println("checkedOut : " + checkedOut + "<br>");
            out.println("------------------------------------------<br>");
            
            for(int i = 0; i < freeConnections.size(); i++) {
                out.println("[" + i + "] " + freeConnections.elementAt(i) + "<br>");
            }
        }
        public void print(PrintStream out) {
            out.print("------------------------------------------\n");
            out.print("name       : " + name + "\n");
            out.print("url        : " + URL + "\n");
            out.print("user       : " + user + "\n");
            out.print("password   : " + password + "\n");
            out.print("maxconn    : " + maxConn + "\n");
            out.print("initconn   : " + initConn + "\n");
            out.print("checkedOut : " + checkedOut + "\n");
            out.print("------------------------------------------\n");
            
            for(int i = 0; i < freeConnections.size(); i++) {
                out.print("[" + i + "] " + freeConnections.elementAt(i) + "\n");
            }
        }
    }
} 
